from typing import Any, Callable, Generator, Self, overload

import numpy as np
import typeguard
from numpy.typing import DTypeLike

from . import _labels, _vector


@typeguard.typechecked
class WriteJagged:
    """
    Write numpy arrays with one variable length axis into a memory mapped file.
    The underlying file is a vector of the concatenated arrays. The index of
    the start of each array is stored in a second vector. This allows for fast
    random access of the arrays.
    The resulting memory mapped array is called jagged because along the first
    axis the arrays can have different lengths.

    Attributes
    ---
    vector: _vector.WriteVector
        The vector of the concatenated arrays.

    _index: _vector.WriteVector
        The vector of the start indices of the arrays.
    """

    vector: _vector.WriteVector
    _index: _vector.WriteVector

    def __init__(
        self: Self,
        path: str,
        name: str,
        *,
        dtype: DTypeLike | None = None,
        shape: tuple[int, ...] | None = None,
        check: str | None = None,
    ) -> None:
        """
        Initializes the context manager for writing a arrays with one variable
        length axis into a memory mapped file.

        Parameters
        ---
        path: str
            Path to the directory where the memory mapped file is stored.

        name: str
            Name of the data. This is used as key to retrieve the data via
            `ReadJagged`.

        shape: tuple[int, ...] | None, optional, default: None
            Shape of the data. `None` if the shape is not yet fixed.

        dtype: DTypeLike | None, optional, default: None
            Dtype of the data. `None` if the dtype is not yet fixed.

        check: str | None, optional, default: None
            Checksum of the data. `None` if a timestamp at the time of saving will
            be used as checksum.
        """
        self.vector = _vector.WriteVector(
            path,
            name,
            shape=shape,
            dtype=dtype,
            check=check,
        )
        self._index = _vector.WriteVector(
            path,
            _labels.get_index_name(name),
            shape=(),
            dtype=np.int64,
            check=check,
        )

    def __enter__(self: Self) -> Self:
        self.vector.__enter__()
        self._index.__enter__()
        self._index.append(np.array(0, dtype=np.int64))
        return self

    def __exit__(self: Self, *args) -> None:
        self.vector.__exit__(*args)
        self._index.__exit__(*args)

    def append(self: Self, value: np.ndarray) -> None:
        """
        Append a new array to the memory mapped file. The first axis of the
        given array is treated as the variable length axis. All other axes are
        fixed by the first call of this method.

        Parameters
        ---
        value: np.ndarray
            The array to append.

        Raises
        ---
        ValueError
            If dtype or all but the first axis of the given array do not match
            the previously appended arrays.
        """
        self.vector.extend(value)
        self._index.append(np.array(self.vector._index, dtype=np.int64))


@typeguard.typechecked
class ReadJagged:
    """
    Read a memory mapped array of arrays with one variable length axis, i.e. a
    memory mapped, jagged array. This array can be indexed or iterated over.

    Attributes
    ---
    vector: np.memmap
        The memory mapped array of the concatenated arrays.

    _index: np.memmap
        The memory mapped array of the start indices of the arrays.
    """

    vector: np.memmap
    _index: np.memmap

    def __init__(self: Self, path: str, name: str) -> None:
        """
        Loads a memory mapped array of arrays with one variable length axis,
        i.e. a memory mapped, jagged array.

        Parameters
        ---
        path: str
            Path to the directory where the memory mapped file is stored.

        name: str
            Name of the data. This name was used as a key to store the data via
            the `WriteJagged` context manager.

        Raises
        ---
        KeyError
            If the name is not found in the metadata.

        FileNotFoundError
            If the file for the name is not found. This could be because the
            data was not saved properly, or the user deleted/moved/renamed the
            underlying file.

        TypeError
            If any arguments python type does not match the type hints.
        """
        self.vector = _vector.read_vector(path, name)
        self._index = _vector.read_vector(path, _labels.get_index_name(name))

    @overload
    def __getitem__(self: Self, index: int) -> np.memmap:
        ...

    @overload
    def __getitem__(self: Self, index: np.ndarray) -> list[np.memmap]:
        ...

    def __getitem__(
        self: Self,
        index: int | np.ndarray,
    ) -> np.memmap | list[np.memmap]:
        """
        Index the jagged array. Indexing with an integer `i` returns the `i`-th
        array. Indexing with an array of integers returns a list of the
        corresponding arrays.

        Parameters
        ---
        index: int | np.ndarray
            The index or indices to retrieve.

        Returns
        ---
        np.memmap | list[np.memmap]
            The retrieved array or list of arrays.

        Raises
        ---
        IndexError
            If the index is out of bounds.

        ValueError
            If the index is not valid, i.e. not positive or one-dimensional.

        TypeError
            If any arguments python type does not match the type hints.
        """
        if isinstance(index, int):
            index = np.array([index])
            return self[index][0]

        assert isinstance(index, np.ndarray)

        if index.ndim == 0:
            error = "Index must be one dimensional."
            raise ValueError(error)

        if index.dtype == bool:
            index = np.where(index)[0]

        if np.any(index < 0):
            error = "Index must be positive."
            raise ValueError(error)

        if index.max() >= len(self._index) - 1:
            error = "Index out of bounds."
            raise IndexError(error)

        indicies = zip(self._index[index], self._index[index + 1])

        # preallocate result, this should be a bit faster
        result = [None] * len(index)

        for i, (s, e) in enumerate(indicies):
            result[i] = self.vector[s:e]  # type: ignore

        return result  # type: ignore

    def __len__(self) -> int:
        """
        The length of the jagged array, i.e. how often the `append` method was
        called on the corresponding `WriteJagged` context manager.

        Returns
        ---
        int
            The number of subarrays. in the jagged array.
        """
        return len(self._index) - 1

    def __iter__(self) -> Generator[np.memmap, None, None]:
        """
        Iterate over the jagged array, i.e. iterate over the subarrays.

        Returns
        ---
        Generator[np.memmap, None, None]
            A generator yielding the subarrays as `np.memmap`s.
        """
        for i in range(len(self)):
            yield self[i]


@typeguard.typechecked
class WriteShaped:
    """
    Write numpy arrays with arbitrary shapes into a memory mapped file. The
    underlying data is stored in a jagged array of flattened arrays. The shapes
    are stored in a second jagged array.

    Attributes
    ---
    vector: WriteJagged
        The jagged array of flattened arrays.

    _shape: WriteJagged
        The jagged array of shapes.
    """

    vector: WriteJagged
    _shape: WriteJagged

    def __init__(
        self: Self,
        path: str,
        name: str,
        *,
        dtype: DTypeLike | None = None,
        check: str | None = None,
    ) -> None:
        """
        Initializes the context manager for writing a arrays with arbitrary
        shapes into a memory mapped file.

        Parameters
        ---
        path: str
            Path to the directory where the memory mapped file is stored.

        name: str
            Name of the data. This is used as key to retrieve the data via
            `ReadShaped`.

        dtype: DTypeLike | None, optional, default: None
            Dtype of the data. `None` if the dtype is inferred from the first
            call of `append`.

        check: str | None, optional, default: None
            Checksum of the data. `None` if a timestamp at the time of saving
            will be used as checksum.

        Raises
        ---
        TypeError
            If any arguments python type does not match the type hints.
        """
        self.vector = WriteJagged(
            path,
            name,
            shape=(),
            dtype=dtype,
            check=check,
        )
        self._shape = WriteJagged(
            path,
            _labels.get_shape_name(name),
            shape=(),
            dtype=np.int64,
        )

    def __enter__(self: Self) -> Self:
        self.vector.__enter__()
        self._shape.__enter__()
        return self

    def __exit__(self: Self, *args) -> None:
        self.vector.__exit__(*args)
        self._shape.__exit__(*args)

    def append(self: Self, value: np.ndarray) -> None:
        """
        Append a new array to the memory mapped file. The dtype of the given
        array must match the dtype of the previously appended arrays.

        Parameters
        ---
        value: np.ndarray
            The array to append.

        Raises
        ---
        ValueError
            If dtype of the given array does not match the previously appended
            arrays.

        TypeError
            If any arguments python type does not match the type hints.
        """
        self.vector.append(value.reshape(-1))
        self._shape.append(np.array(value.shape, dtype=np.int64))


@typeguard.typechecked
class ReadShaped:
    """
    Read a memory mapped array of arrays with arbitrary shapes. This array can
    be indexed or iterated over.

    Attributes
    ---
    vector: ReadJagged
        The jagged array of flattened arrays.

    _shape: ReadJagged
        The jagged array of shapes.
    """

    vector: ReadJagged
    _shape: ReadJagged

    def __init__(self: Self, path: str, name: str) -> None:
        """
        Loads a memory mapped array of arrays with arbitrary shapes.

        Parameters
        ---
        path: str
            Path to the directory where the memory mapped file is stored.

        name: str
            Name of the data. This name was used as a key to store the data via
            the `WriteShaped` context manager.

        Raises
        ---
        KeyError
            If the name is not found in the metadata.

        FileNotFoundError
            If the file for the name is not found. This could be because the
            data was not saved properly, or the user deleted/moved/renamed the
            underlying file.

        TypeError
            If any arguments python type does not match the type hints.
        """
        self.vector = ReadJagged(path, name)
        self._shape = ReadJagged(path, _labels.get_shape_name(name))

    @overload
    def __getitem__(self: Self, index: int) -> np.memmap:
        ...

    @overload
    def __getitem__(self: Self, index: np.ndarray) -> list[np.memmap]:
        ...

    def __getitem__(
        self: Self,
        index: int | np.ndarray,
    ) -> np.memmap | list[np.memmap]:
        """
        Get the array(s) at the given index or indices.

        Parameters
        ---
        index: int | np.ndarray
            The index or indices to retrieve.

        Returns
        ---
        np.memmap | list[np.memmap]
            If an integer is given, the array at the given index is returned.
            If an array of integers is given, a list of the corresponding
            arrays is returned.

        Raises
        ---
        IndexError
            If the index is out of bounds.

        ValueError
            If the index is not valid, i.e. not positive or one-dimensional.

        TypeError
            If any arguments python type does not match the type hints.
        """
        if isinstance(index, int):
            index = np.array([index])
            return self[index][0]

        assert isinstance(index, np.ndarray)

        if index.ndim == 0:
            error = "Index must be one dimensional."
            raise ValueError(error)

        if index.dtype == bool:
            index = np.where(index)[0]

        if np.any(index < 0):
            error = "Index must be positive."
            raise ValueError(error)

        shape = self._shape[index]
        arrays = self.vector[index]

        for i, array in enumerate(arrays):
            arrays[i] = array.reshape(shape[i])  # type: ignore

        return arrays

    def __len__(self: Self) -> int:
        """
        Get the length of the array of heterogeneous arrays, i.e. how often the
        `append` method was called on the corresponding `WriteShaped` context
        manager.

        Returns
        ---
        int
            The number of arrays in the array of heterogeneous arrays.
        """
        return len(self.vector)

    def __iter__(self: Self) -> Generator[np.memmap, None, None]:
        """
        Iterate over the array of heterogeneous arrays.

        Returns
        ---
        Generator[np.ndarray, None, None]
            A generator yielding the arrays as `np.ndarray`s.
        """
        for i in range(len(self)):
            yield self[i]


@typeguard.typechecked
class WriteStructured:
    """
    Write arbitrary structured data into a memory mapped file.

    Attributes
    ---
    data: WriteShaped
        Where the flattened data is stored.

    structure: WriteShaped
        Where the structure of the data is stored.

    get_structure: Callable[[np.ndarray], np.ndarray]
        A function that takes an array and returns its structure. This structure
        will later be used in the `ReadStructured` class to reconstruct the
        data.

    """

    get_structure: Callable[[np.ndarray], np.ndarray]
    data: WriteShaped
    structure: WriteShaped

    def __init__(
        self: Self,
        path: str,
        name: str,
        *,
        get_structure: Callable[[np.ndarray], np.ndarray],
        check: str | None = None,
    ) -> None:
        """
        Initializes the context manager for writing a arrays with arbitrary
        structure into a memory mapped file.

        Parameters
        ---
        path: str
            Path to the directory where the memory mapped file is stored.

        name: str
            Name of the data. This is used as key to retrieve the data via
            `ReadStructured`.

        get_structure: Callable[[np.ndarray], np.ndarray]
            A function that takes an array and returns its structure. This
            structure will later be used in the `ReadStructured` class to
            reconstruct the data.

        check: str | None, optional, default: None
            Checksum of the data. `None` if a timestamp at the time of saving
            will be used as checksum.

        Raises
        ---
        TypeError
            If any arguments python type does not match the type hints.
        """
        self.get_structure = get_structure
        self.data = WriteShaped(path, name, check=check)
        self.structure = WriteShaped(
            path,
            _labels.get_structured_name(name),
            check=check,
        )

    def __enter__(self: Self) -> Self:
        self.data.__enter__()
        self.structure.__enter__()
        return self

    def __exit__(self: Self, *args) -> None:
        self.data.__exit__(*args)
        self.structure.__exit__(*args)

    def append(self, value: np.ndarray) -> None:
        """
        Append a new array to the memory mapped file. The dtype of the given
        array must match the dtype of the previously appended arrays.

        Parameters
        ---
        value: np.ndarray
            The array to append.

        Raises
        ---
        ValueError
            If dtype of the given array does not match the previously appended
            arrays.

        TypeError
            If any arguments python type does not match the type hints.
        """
        self.data.append(value)
        self.structure.append(self.get_structure(value))


@typeguard.typechecked
class ReadStructured:
    """
    Read a memory mapped array of arrays with arbitrary structure. This array
    can be indexed or iterated over.

    Attributes
    ---
    reconstruct: Callable[[np.ndarray, np.ndarray], Any]
        A function that takes an array and its structure and returns the
        original data.

    data: ReadShaped
        Where the flattened data is stored.

    structure: ReadShaped
        Where the structure of the data is stored.
    """

    reconstruct: Callable[[np.ndarray, np.ndarray], Any]
    data: ReadShaped
    structure: ReadShaped

    def __init__(
        self: Self,
        path: str,
        name: str,
        reconstruct: Callable[[np.ndarray, np.ndarray], Any],
    ) -> None:
        """
        Loads a memory mapped array of arrays with arbitrary structure.

        Parameters
        ---
        path: str
            Path to the directory where the memory mapped file is stored.

        name: str
            Name of the data. This name was used as a key to store the data via
            the `WriteStructured` context manager.

        reconstruct: Callable[[np.ndarray, np.ndarray], Any]
            A function that takes an array and its structure and returns the
            original data.

        Raises
        ---
        KeyError
            If the name is not found in the metadata.

        FileNotFoundError
            If the file for the name is not found. This could be because the
            data was not saved properly, or the user deleted/moved/renamed the
            underlying file.

        TypeError
            If any arguments python type does not match the type hints.
        """
        self.reconstruct = reconstruct
        self.data = ReadShaped(path, name)
        self.structure = ReadShaped(path, _labels.get_structured_name(name))

    @overload
    def __getitem__(self: Self, index: int) -> Any:
        ...

    @overload
    def __getitem__(self: Self, index: np.ndarray) -> list[Any]:
        ...

    def __getitem__(self: Self, index: int | np.ndarray) -> Any | list[Any]:
        """
        Index the array of arrays with arbitrary structure. Indexing with an
        integer `i` returns the `i`-th array. Indexing with an array of
        integers returns a list of the corresponding arrays.

        Parameters
        ---
        index: int | np.ndarray
            The index or indices to retrieve.

        Returns
        ---
        Any | list[Any]
            The retrieved array or list of arrays.

        Raises
        ---
        IndexError
            If the index is out of bounds.

        ValueError
            If the index is not valid, i.e. not positive or one-dimensional.

        TypeError
            If any arguments python type does not match the type hints.
        """
        if isinstance(index, int):
            index = np.array([index])
            return self[index][0]

        data = self.data[index]
        structure = self.structure[index]

        for i, (d, s) in enumerate(zip(data, structure)):
            data[i] = self.reconstruct(d, s)

        return data



import collections.abc
import math
from typing import (
    Any,
    Callable,
    Hashable,
    Iterator,
    Literal,
    Mapping,
    Protocol,
    Self,
    overload,
    runtime_checkable,
)

import numpy as np
import typeguard

__all__ = [
    "SequenceDict",
    "unwrap_recursively",
    "Sliced",
    "Subindexed",
    "FastShuffled",
    "Batched",
    "split",
]

# TODO implement kfold, and mapping, also norm?


@runtime_checkable
class NumpySequence(Protocol):
    """
    Similar to collections.abc.Sequence, but also allows for indexing with
    numpy arrays, to get multiple elements at once.
    """

    @overload
    def __getitem__(self: Self, index: int) -> Any:
        ...

    @overload
    def __getitem__(self: Self, index: slice) -> "NumpySequence":
        ...

    @overload
    def __getitem__(self: Self, index: np.ndarray) -> "NumpySequence":
        ...

    def __getitem__(self: Self, index: int | slice | np.ndarray) -> Any:
        ...

    def __iter__(self: Self) -> Iterator:
        ...

    def __len__(self: Self) -> int:
        ...


def unwrap_recursively(x):
    """
    Try to use the `unwrap` method of an object recursively, if it exists,
    otherwise return the object itself.
    This is usefull for getting the underlying data out of lazy sequence
    manipulations.
    """
    if hasattr(x, "unwrap"):
        return unwrap_recursively(x.unwrap())

    return x


@typeguard.typechecked
class SequenceDict(collections.abc.Sequence):
    """
    Comnine multiple sequences into a single sequence using a dictionary.
    """

    def __init__(
        self: Self,
        _sequences: dict,
    ) -> None:
        self._sequences = _sequences

        if len(set(map(len, _sequences.values()))) != 1:
            raise ValueError("All sequences must have the same length.")

        self._length = len(next(iter(_sequences.values())))

    @overload
    def __getitem__(self: Self, index: int) -> dict[Hashable, Any]:
        ...

    @overload
    def __getitem__(self: Self, index: slice) -> "NumpySequence":
        ...

    @overload
    def __getitem__(self: Self, index: np.ndarray) -> "NumpySequence":
        ...

    def __getitem__(
        self: Self,
        index: int | slice | np.ndarray,
    ) -> Any:
        if isinstance(index, int):
            if index < 0:
                index += len(self)

            if index < 0 or index >= len(self):
                raise IndexError(f"Index {index} out of bounds.")

            return {k: v[index] for k, v in self._sequences.items()}

        indexed = {}
        for k, v in self._sequences.items():
            indexed[k] = v[index]

        return SequenceDict(indexed)

    def __len__(self: Self) -> int:
        return self._length

    def unwrap(
        self: Self,
    ) -> Mapping[Hashable, NumpySequence | np.memmap | np.ndarray]:
        return self._sequences


@typeguard.typechecked
class _Sliced(collections.abc.Sequence):
    """
    Lalazily slice of a sequence by wrapping the __getitem__ method.
    """

    def __init__(
        self,
        _sequence: NumpySequence | np.memmap | np.ndarray,
        _slice: slice,
        /,
    ) -> None:
        self._sequence = _sequence

        self._start = _slice.start or 0
        self._stop = _slice.stop or len(_sequence)
        self._step = _slice.step or 1

        if self._start < 0:
            self._start += len(_sequence)

        if self._stop < 0:
            self._stop += len(_sequence)

        # check if start or stop are out of bounds
        if not 0 <= self._start < len(_sequence):
            raise IndexError(f"Start index {self._start} out of bounds.")

        if not 0 <= self._stop <= len(_sequence):
            raise IndexError(f"Stop index {self._stop} out of bounds.")

    def __getitem__(self: Self, index: int | slice | np.ndarray) -> Any:
        if isinstance(index, int):
            if index < 0:
                index += len(self)

            if index < 0 or index >= len(self):
                raise IndexError(f"Index {index} out of bounds.")

            return self._sequence[self._start + self._step * index]

        if isinstance(index, slice):
            return Sliced(self, index)

        if isinstance(index, np.ndarray):
            index[index < 0] += len(self)

            if any(index < 0) or any(index >= len(self)):
                raise IndexError(f"Index {index} out of bounds.")

            return self._sequence[self._start + self._step * index]

        raise TypeError(f"Invalid index type: {type(index)}")

    def __len__(self: Self) -> int:
        return (self._stop - self._start + self._step - 1) // self._step

    def unwrap(self: Self) -> NumpySequence | np.memmap | np.ndarray:
        index = np.arange(len(self))
        return self[index]


def Sliced(
    x: NumpySequence | np.memmap | np.ndarray,
    s: slice,
) -> _Sliced | np.memmap | np.ndarray:
    """
    Slice a sequence. If it is a numpy array, the slicing is done directly,
    otherwise a lazy wrapper is returned.
    """
    if isinstance(x, (np.ndarray, np.memmap, list, tuple, range)):
        return x[s]

    return _Sliced(x, s)


@typeguard.typechecked
class Subindexed(collections.abc.Sequence):
    """
    Restrict a sequence to a subset of indicies.
    """

    def __init__(
        self: Self,
        _sequence: NumpySequence | np.memmap | np.ndarray,
        _subindex: np.ndarray | np.memmap,  # TODO also a NumpySequence (?)
    ) -> None:
        if _subindex.ndim != 1:
            raise ValueError("Subindex must have ndim=1.")

        if not np.issubdtype(_subindex.dtype, np.integer):
            raise ValueError("Subindex must have integer dtype.")

        # do not test whether subindex is in bounds, the sequence and
        # subindex might be very large and only small parts of this index
        # might be used, so the wait might be much longer than the actual
        # time this index would be used for
        self._sequence = _sequence
        self._subindex = _subindex.astype(np.intp)

    def __getitem__(self, index: int | slice | np.ndarray) -> Any:
        if isinstance(index, int):
            return self._sequence[self._subindex[index]]

        if isinstance(index, slice):
            return Subindexed(self._sequence, self._subindex[index])

        if isinstance(index, np.ndarray):
            # print(index)
            return self._sequence[self._subindex[index]]

        raise TypeError(f"Invalid index type: {type(index)}")

    def __len__(self) -> int:
        return len(self._subindex)

    def unwrap(self: Self) -> NumpySequence | np.memmap | np.ndarray:
        print(self._sequence)
        return self._sequence[self._subindex]


@typeguard.typechecked
def permutation_bands(length: int) -> Callable[[np.ndarray], np.ndarray]:
    bands_total = int(math.log2(length) ** 2)
    bands_total = min(bands_total, length // 2)

    band_offsets = np.random.uniform(size=bands_total)
    band_offsets = (band_offsets * length).astype(np.int64)

    @typeguard.typechecked
    def permutation_func(indicies: np.ndarray) -> np.ndarray:
        indicies_offsets = indicies // bands_total
        indicies_remainders = indicies % bands_total

        indicies_offsets += band_offsets[indicies_remainders]
        indicies_offsets %= length // bands_total

        return indicies_offsets * bands_total + indicies_remainders

    return permutation_func


@typeguard.typechecked
def permutation_blocks(
    length: int,
    modulus: int = int(2**17 - 1),
) -> Callable[[np.ndarray], np.ndarray]:
    modulus = min(int(modulus), length)
    subpermutation = np.random.uniform(size=modulus).argsort()

    perm = np.random.uniform(size=modulus)
    perm = np.argsort(perm)

    @typeguard.typechecked
    def permute(indicies: np.ndarray) -> np.ndarray:
        indicies_offsets = indicies // modulus
        indicies_remainders = indicies % modulus

        indicies_remainders = subpermutation[indicies_remainders]

        result = indicies_offsets * modulus + indicies_remainders
        return result % length

    return permute


class FixedSeed:
    def __init__(self, seed: int | None) -> None:
        self._seed = seed

    def __enter__(self: Self) -> None:
        if self._seed is None:
            return

        self.state = np.random.get_state()
        np.random.seed(self._seed)

    def __exit__(self: Self, *_) -> None:
        if self._seed is None:
            return

        np.random.set_state(self.state)


@typeguard.typechecked
def permutation_fast(length: int) -> Callable[[np.ndarray], np.ndarray]:
    perm1 = permutation_blocks(length)
    perm2 = permutation_bands(length)

    @typeguard.typechecked
    def perm(x: np.ndarray) -> np.ndarray:
        x = perm1(x)
        x = perm2(x)
        return x

    return perm


@typeguard.typechecked
class FastShuffled(collections.abc.Sequence):
    """
    Lazily shuffle a sequence using a fast permutation function.
    """

    def __init__(
        self,
        sequence: NumpySequence | np.memmap | np.ndarray,
        /,
        *,
        seed: int | None = None,
    ) -> None:
        self._sequence = sequence

        with FixedSeed(seed):
            self._permutation = permutation_fast(len(sequence))

    def __getitem__(self, index: int | slice | np.ndarray) -> Any:
        if isinstance(index, int):
            if index < 0:
                index += len(self)

            if index < 0 or index >= len(self):
                raise IndexError(f"Index {index} out of bounds.")

            index_arr = np.array([index])
            index_int = self._permutation(index_arr)[0]
            return self._sequence[index_int]

        if isinstance(index, slice):
            return Sliced(self, index)

        if isinstance(index, np.ndarray):
            index[index < 0] += len(self)

            if any(index < 0) or any(index >= len(self)):
                raise IndexError(f"Index {index} out of bounds.")

            return self._sequence[self._permutation(index)]

        raise TypeError(f"Invalid index type: {type(index)}")

    def __len__(self) -> int:
        return len(self._sequence)

    def unwrap(self) -> NumpySequence | np.memmap | np.ndarray:
        index = np.arange(len(self))
        return self[index]


@typeguard.typechecked
class Batched(collections.abc.Sequence):
    """
    Lazily collect a sequence into batches.
    """

    def __init__(
        self,
        _sequence: NumpySequence | np.memmap | np.ndarray,
        /,
        batch_size: int,
        drop_remainder: bool = False,
    ) -> None:
        self._sequence = _sequence
        self._batch_size = batch_size
        self._drop_remainder = drop_remainder

    def __getitem__(self, index: int | slice | np.ndarray) -> Any:
        if isinstance(index, int):
            if index < 0:
                index += len(self)

            if index < 0 or index >= len(self):
                raise IndexError(f"Index {index} out of bounds.")

            start = index * self._batch_size
            stop = (index + 1) * self._batch_size

            if not self._drop_remainder:
                stop = min(stop, len(self._sequence))

            return self._sequence[start:stop]

        if isinstance(index, slice):
            return Sliced(self, index)

        if isinstance(index, np.ndarray):
            index[index < 0] += len(self)

            if any(index < 0) or any(index >= len(self)):
                raise IndexError(f"Index {index} out of bounds.")

            result = []

            for i in index:
                start = i * self._batch_size
                stop = (i + 1) * self._batch_size

                if not self._drop_remainder:
                    stop = min(stop, len(self._sequence))

                result.append(self._sequence[start:stop])

            return result

        raise TypeError(f"Invalid index type: {type(index)}")

    def __len__(self) -> int:
        if self._drop_remainder:
            return len(self._sequence) // self._batch_size

        return (len(self._sequence) + self._batch_size - 1) // self._batch_size

    def unwrap(self) -> NumpySequence | np.memmap | np.ndarray:
        index = np.arange(len(self))
        return self[index]


@typeguard.typechecked
def split(
    sequence: NumpySequence,
    split_index: int | Literal["train", "valid", "test"] = "train",
    split_fracs: tuple[float, ...] = (0.8, 0.1, 0.1),
    shuffle: bool = True,
    seed: int | None = 42,
) -> _Sliced | np.memmap | np.ndarray:
    """
    Get a slice of a sequence according to a split fraction. This is useful
    for splitting a dataset into training, validation and test set.
    """
    if abs(sum(split_fracs) - 1) > 1e-12:
        raise ValueError("Split fractions must sum to 1.")

    if isinstance(split_index, str):
        split_index = {"train": 0, "valid": 1, "test": 2}[split_index]

    if split_index >= len(split_fracs):
        raise ValueError("Split index out of bounds.")

    if shuffle:
        sequence = FastShuffled(sequence, seed=seed)

    start = int(sum(len(sequence) * f for f in split_fracs[:split_index]))
    length = int(len(sequence) * split_fracs[split_index])
    return Sliced(sequence, slice(start, start + length))


# class Normalize(collections.abc.Sequence):
#     def __init__(self, _sequence: np.ndarray | np.memmap) -> None:
#         self._sequence = _sequence
#         self._mean = _sequence.mean(axis=0, keepdims=True)
#         self._stdd = _sequence.std(axis=0, keepdims=True)

#     def norm(self, x: np.ndarray) -> np.ndarray:
#         return (x - self._mean) / self._stdd

#     def unnorm(self, x: np.ndarray) -> np.ndarray:
#         return x * self._stdd + self._mean

#     def __getitem__(self, index: int | slice | np.ndarray) -> np.ndarray:
#         return self.norm(self._sequence[index])

#     def __len__(self) -> int:
#         return len(self._sequence)

#     def unwrap(self) -> np.ndarray | np.memmap:
#         return self.norm(self._sequence)



try:
    import uproot
except ImportError:
    error = "'Uproot' is a requirement for 'memmpy.reroot', but was not found."
    raise ImportError(error)

import dataclasses
import hashlib
import os
from typing import Any

import awkward as ak
import numpy as np
import tqdm
import typeguard

from . import _labels, _vector

__all__ = [
    "RFileConfig",
    "memmap_root",
]


@dataclasses.dataclass(frozen=True)
class RFileConfig:
    """
    Stores the configuration for a single ROOT file.

    Parameters
    ---
    path: str
        Path to the ROOT file.

    tree: str, optional, default: "nominal_Loose"
        Name of the tree to read from.

    metadata: dict[str, str], optional, default: {}
        Metadata to store in the memmap file. This can be used to store
        information about the what is stored in the file, i.e. from which
        process the data originates, or which run it is from.
    """

    path: str
    tree: str = dataclasses.field(default="nominal_Loose")
    metadata: dict[str, str] = dataclasses.field(default_factory=dict)

    def __lt__(self, other: "RFileConfig") -> bool:
        return repr(self) < repr(other)


def _encode_metadata(
    metadata: dict[str, str],
    size: int,
) -> dict[str, np.ndarray]:
    metadata_encoded = {}

    for k, v in metadata.items():
        hashed = _labels.hash_64bit(v.encode())
        metadata_encoded[k] = np.full(
            size,
            hashed,
            dtype="uint64",
        )

    return metadata_encoded


@typeguard.typechecked
def memmap_root(
    root_files: list[RFileConfig],
    path_mmap: str,
    keys: set[str],
    keys_padded: dict[str, tuple[int, Any]] | None = None,
    aliases: dict[str, str] | None = None,
    recreate: bool = False,
) -> dict[str, np.memmap]:
    """
    Store the data from the given ROOT files in a memmap files.

    Parameters
    ---
    root_files: list[RFileConfig]
        List of ROOT file configurations to read from, i.e. the path to the file,
        the name of the tree to read from, and metadata to store in memmaps.

    path_mmap: str
        Path to the memmap directory.

    keys: set[str]
        Set of keys to read from the ROOT files. The keys must be scalar valued.

    keys_padded: dict[str, tuple[int, Any]], optional, default: None
        Dictionary of 1d vector valued keys to read from the ROOT files. The
        dictionary maps the key to a tuple of the size of the vector and the
        value to pad with.

    aliases: dict[str, str], optional, default: None
        Dictionary of aliases to use for the keys.

    recreate: bool, optional, default: False
        If True, the memmap files are recreated even if they already exist.

    Raises
    ---
    FileNotFoundError
        If a ROOT file is not found.

    TypeError
        If the arguments do not match the type hints.
    """
    keys_padded = keys_padded or {}
    aliases = aliases or {}
    keys = keys | set(keys_padded.keys())

    hasher = hashlib.sha256()
    for k in sorted(aliases.keys()):
        hasher.update(str(k).encode())
        hasher.update(aliases[k].encode())

    for rfile in sorted(root_files):
        hasher.update(os.path.basename(rfile.path).encode())
        hasher.update(str(os.path.getmtime(rfile.path)).encode())
        hasher.update(str(os.path.getsize(rfile.path)).encode())
        hasher.update(str(rfile.tree).encode())
        hasher.update(str(rfile.metadata).encode())

    hashed = hasher.hexdigest()

    metadata_values = set()
    metadata_keys = set()

    for rfile in root_files:
        metadata_values |= set(rfile.metadata.values())
        metadata_keys |= set(rfile.metadata.keys())

    if not recreate:
        meta = _labels.safe_load(path_mmap)

        for k in keys:
            if k not in meta["arrays"]:
                break

            if meta["arrays"][k]["check"] != hashed:
                break

            if k in keys_padded:
                if meta["arrays"][k]["shape"][-1] != keys_padded[k][0]:
                    break

        else:
            path_mmap = os.path.abspath(path_mmap)
            print(f"Using cached files in '{path_mmap}' for {list(keys)}.")
            return _vector.read_vectors(path_mmap, keys | metadata_keys)

    with _vector.WriteVectorDict(path_mmap, check=hashed) as vectors:
        for rfile in tqdm.tqdm(root_files):
            tqdm.tqdm.write(f"processing '{rfile.path}'")

            if not os.path.exists(rfile.path):
                raise FileNotFoundError(f"File '{rfile.path}' not found.")

            with uproot.open(rfile.path) as file:  # type: ignore
                data_root = file[rfile.tree].arrays(  # type: ignore
                    keys,
                    aliases=aliases,
                    library="ak",
                    how=dict,
                )
                data_numpy = {}

                for k, v in data_root.items():
                    if k in keys_padded:
                        assert v.ndim == 2
                        pad_size, pad_value = keys_padded[k]
                        array = ak.pad_none(v, pad_size, clip=True)
                        array = ak.fill_none(array, pad_value)
                        data_numpy[k] = array.to_numpy()  # type: ignore
                        continue

                    assert v.ndim == 1
                    data_numpy[k] = v.to_numpy()

                size = next(iter(data_numpy.values())).shape[0]

                data_path = _encode_metadata(rfile.metadata, size)
                vectors.expand(data_numpy | data_path)

        print("Saving...")

    # add hahses to metadata
    for value in metadata_values:
        _labels.add_hash(path_mmap, value)

    print("Saving done.")
    return _vector.read_vectors(path_mmap, keys | metadata_keys)



from ._labels import hash_64bit
from ._loader import (
    Batched,
    FastShuffled,
    SequenceDict,
    Sliced,
    Subindexed,
    split,
    unwrap_recursively,
)
from ._subset import compute_cut_batched
from ._vector import WriteVector, WriteVectorDict, read_vector

__all__ = [
    "hash_64bit",
    "Batched",
    "FastShuffled",
    "SequenceDict",
    "Sliced",
    "Subindexed",
    "split",
    "unwrap_recursively",
    "compute_cut_batched",
    "WriteVector",
    "WriteVectorDict",
    "read_vector",
]

__version__ = "0.1.5"
